import re
import warnings

from flap.utils import repl_positions


def is_postcode(string):
    postcode_regex = r'^([A-Z]{1,2})([0-9][A-Z0-9]?) ?([0-9][A-Z]{2})$'
    if isinstance(string, str):
        return bool(re.match(postcode_regex, string))
    else:
        return False


def postcode_preproc(string, suppress_warnings=False):
    postcode_regex = r'^([A-Z]{1,2})([0-9][A-Z0-9]?) ?([0-9][A-Z]{2})$'

    if is_postcode(string):
        return re.sub(pattern=postcode_regex, repl=r'\1\2 \3', string=string)
    else:
        if not suppress_warnings:
            warnings.warn('%s is not a postcode' % string)
        return string


def address_line_preproc(address):
    preproc_mapping = [
        (r"[^\w\-/\s'\.\(\)]", ' '),
        (r'  +', ' '),
        (r'--+', '-'),
        (r'//+', '/')
    ]

    n_sub = 0
    n_abbr = 0
    add_pt = 0
    add_dl = 0
    remove_region = 0

    for pattern, repl in preproc_mapping:
        address, n = re.subn(pattern=pattern, repl=repl, string=address)
        n_sub += n

    address = address.upper()

    log = {'address_preproc': address, 'preproc_n_sub': n_sub, 'preproc_n_abbr': n_abbr, 'add_pt': add_pt,
           'add_dl': add_dl, 'remove_region': remove_region}

    return address, log



"""

Sql based match using combined queries (Need an evaluation for query efficiency)

Accelerate with indexing sql db

Accelerate with in-memory db

Linear alignment for each results

Avoid duplicated alignment using a memory dict

Use a classifier for early stop


Logic flow of matcher

require (parser object, db object, address, scorer, query priority, threshold score, max beam_width)

input (parsed address, db)

parse address ->

while < threshold and < max beam width

    query next rules

    update current beam width

    for row in query results

        match and score

get match of max score


To do:

Add functionality for multiplier addresses
    - [x] Index data
    - [ ] Index database
    - [x] Creating master file for scotland


Adding more rules and studying effectiveness of rules

How multipliers should be applied

If multiplier regex found in address

    If multiplier in UPRN
        Just match

    If multiplier not in UPRN
        guess the multiplier value
        convert the multiplier value to normal value

        get best match based on the model

Guess multiplier with UPRN information


"""
import os

import pandas as pd
import re
import numpy as np
import traceback
import pickle

from scipy.spatial.distance import pdist

import flap

from flap.database.sql import SqlDB
from flap.parser.rule_parser_fast import RuleParserFast
from flap.preprocessing import postcode_preproc, address_line_preproc

from flap.alignment.linear_assignment_alignment import LinearAssignmentAlignment
from flap.utils.progress_map import progress_map_tqdm_concurrent

# import time  # time

MODULE_PATH = os.path.dirname(flap.__file__)

try:
    DEFAULT_MODEL_PATH = [os.path.join(MODULE_PATH, 'model', path)
                          for path in os.listdir(os.path.join(MODULE_PATH, 'model')) if 'clf' in path][0]
except FileNotFoundError:
    DEFAULT_MODEL_PATH = None

DEFAULT_MULTIPLIER_INDEX = os.path.join(MODULE_PATH, 'parser', 'multiplier_index.csv')


class SqlMatcher:

    queries = [
        ['POSTCODE'],
        ['pc_area', 'pc_district', 'number_like_0'],
        ['POST_TOWN', 'THOROUGHFARE'],
        ['POST_TOWN', 'number_like_0']
    ]

    def __init__(self, sql_db, scorer=None, multiplier_indices=None):
        self.sql_db = sql_db
        self.parser = RuleParserFast(sql_db)
        self.scorer = scorer
        self.multiplier_indices = multiplier_indices

        if scorer is None:
            if DEFAULT_MODEL_PATH is not None:
                try:
                    self.scorer = ClassifierScorer(DEFAULT_MODEL_PATH)
                except FileNotFoundError:
                    pass

        if multiplier_indices is None:
            try:
                self.multiplier_indices = pd.read_csv(DEFAULT_MULTIPLIER_INDEX, index_col=0)
            except FileNotFoundError:
                pass

        for query in self.queries:
            self.sql_db.create_index(table_name='indexed', columns=query)

#         self.t_temp = time.time()  # time
#         self.t_log = {}  # time

#     def _record_time(self, checkpoint_name):  # time
#         # time
#         self.t_log[checkpoint_name] = time.time() - self.t_temp  # time
#         self.t_temp = time.time()  # time

    def _query(self, query, parsed):

        query_task = {k: parsed['FOR_QUERY'][k].replace("'", "''") for k in query}

        query_string = query_task_to_sql(query_task)
        res = pd.DataFrame(self.sql_db.sql_query(query_string), columns=self.sql_db.get_columns_of_table('indexed'))

        return res

    @staticmethod
    def _match_one_record(parsed, uprn_row):

        uprn_prepared = prepare_uprn(uprn_row)

        number_like_matching_matrix = get_number_like_matching_matrix(list(parsed['NUMBER_LIKE'].values()),
                                                                      uprn_prepared['NUMBER_LIKE'])

        ar = LinearAssignmentAlignment(parsed['TEXTUAL'], uprn_prepared['TEXTUAL']).get_result()

        text_alignment_score = ar.get_score()

        try:
            postcode_matching_vector = postcode_matching(parsed['POSTCODE_SPLIT'], uprn_prepared['POSTCODE_SPLIT'])
        except KeyError:
            postcode_matching_vector = [0] * 5

        x = summarize_features(number_like_matching_matrix, text_alignment_score, postcode_matching_vector)

        return {'uprn_row': uprn_row, 'features': x}

    def generate_feature(self, address, uprn_row):

        address, _ = address_line_preproc(address)

        parsed = self.parser.parse(address, method='fast')

        res = self._match_one_record(parsed, uprn_row)

        return res['features']

    def match(self, address, max_beam_width=500, score_threshold=0.3, troubleshoot=False):

        #         self.t_temp = time.time()  # time
        #         self._record_time('start')  # time

        address, log = address_line_preproc(address)

        if not log['proceed']:
            return {'score': 0}

        #         self._record_time('preproc')  # time

        parsed = self.parser.parse(address, method='fast')

        #         self._record_time('fast_parse')  # time

        current_bw = 0

        stop_loop = False

        best_res = None

        results = []

        for query in self.queries:

            # print(query)
            try:
                res = self._query(query, parsed)

            except KeyError:
                parsed = self.parser.parse(address, method='all')
                try:
                    res = self._query(query, parsed)
                except KeyError:
                    continue

            current_bw += len(res)

            # Matching loop

            for i, row in res.iterrows():
                results.append(self._match_one_record(parsed, row))

            # Deal with multiplier

            multi_regex = re.compile(r'(\d)[A-Z](\d+)')

            multi_regex_matches = {k: re.match(multi_regex, v) for k, v in parsed['NUMBER_LIKE'].items()}

            if any([multi_regex_matches[k] is not None for k in multi_regex_matches]):

                alt_address = address

                alt_parsed = parsed.copy()

                number_like = alt_parsed['NUMBER_LIKE']

                level = 1
                key = 'number_like_0'
                flat = 1

                list_of_indices = []

                for k, m in multi_regex_matches.items():
                    if m is None:
                        list_of_indices.append(number_like[k])
                    else:
                        level = int(m.group(1))
                        flat = int(m.group(2))
                        key = k
                        break

                if level == 1:
                    alt_parsed['NUMBER_LIKE'][key] = str(flat)
                    alt_address = re.sub(pattern=parsed['FOR_QUERY'][key], repl=alt_parsed['NUMBER_LIKE'][key],
                                         string=alt_address)

                    alt_parsed['TEXTUAL'] = strip_number_like(alt_address)

                    for i, row in res.iterrows():
                        results.append(self._match_one_record(alt_parsed, row))

                    continue

                if len(list_of_indices) == 0:
                    continue

                if self.multiplier_indices is not None:

                    max_level = 0
                    max_flat = 0

                    try:
                        index = '--'.join(['-'.join(list_of_indices), parsed['FOR_QUERY']['POSTCODE']])

                        if index in self.multiplier_indices.index:
                            max_level = self.multiplier_indices.loc[index, 'max_level']
                            max_flat = self.multiplier_indices.loc[index, 'max_flat']

                    except KeyError:

                        pass

                    for i, row in res.iterrows():

                        try:
                            n_tenement = int(row['n_tenement'])
                        except ValueError:
                            n_tenement = 1

                        multiplier, _ = guess_multiplier(max_level, max_flat, n_tenement)
                        alt_number = (level - 1) * multiplier + flat
                        alt_parsed['NUMBER_LIKE'][key] = str(alt_number)

                        alt_address = re.sub(pattern=parsed['FOR_QUERY'][key], repl=alt_parsed['NUMBER_LIKE'][key],
                                             string=alt_address)
                        alt_parsed['TEXTUAL'] = strip_number_like(alt_address)

                        results.append(self._match_one_record(alt_parsed, row))

            # Deal with Special

            number_like = list(parsed['NUMBER_LIKE'].values())
            if 'GROUND' in number_like:
                the_index = 'number_like_%s' % number_like.index('GROUND')
                alt_parsed = parsed.copy()
                alt_parsed['NUMBER_LIKE'][the_index] = '0'

                for i, row in res.iterrows():
                    results.append(self._match_one_record(alt_parsed, row))

            # Score and select

            try:
                if self.scorer is not None:
                    X = [res['features'] for res in results]
                    #                     t0 = time.time() # time
                    scores = self.scorer.score_batch(X)
                    #                     self.t_log['scorer'] = time.time() - t0 # time
                    for res, score in zip(results, scores):
                        res['score'] = score

                best_res = max(results, key=lambda x: x['score'])
                if best_res['score'] > score_threshold:
                    stop_loop = True

            except ValueError:
                pass

            if current_bw > max_beam_width:
                stop_loop = True

            if stop_loop:
                break

        # End the main loop

        #         self._record_time('main_loop')  # time
        #         self.t_log['beam_width'] = current_bw  # time

        if best_res is None:
            best_res = {'score': 0}

        if troubleshoot:
            return best_res, results
        else:
            return best_res

    def _match_safe(self, address, max_beam_width=500, score_threshold=0.3, troubleshoot=False):

        try:
            return self.match(address, max_beam_width, score_threshold, troubleshoot)
        except:
            try:
                return {'score': 0, 'error': '\n'.join([address, traceback.format_exc()])}
            except:
                return {'score': 0, 'error': traceback.format_exc()}

    def match_batch(self, list_of_addresses, max_workers=None, chunksize=None):

        results = progress_map_tqdm_concurrent(self._match_safe, list_of_addresses,
                                               max_workers=max_workers, chunksize=chunksize)

        return results


class RuleScorer:

    """
    Rule scorer takes in a vector and return one scalar score
    """

    @staticmethod
    def weighted_score(x):
        x = np.array(x)
        w = np.array([5, 4, 3, 2, 1, 4, 5, 3, 2, 1, 3, 4, 5, 2, 1, 2, 3, 4, 5, 1, 1, 2, 3, 4, 5,
                      0, 0, -0.01, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                      1, 1, 1, 1, 1])
        return sum(np.multiply(x, w))

    def __init__(self):
        pass

    def score(self, x, method='weighted_score'):
        score_func = self.__getattribute__(method)
        return score_func(x)

    def score_batch(self, X, method='weighted_score'):
        score_func = self.__getattribute__(method)
        return [score_func(x) for x in X]


class ClassifierScorer:

    def __init__(self, model_path):
        with open(model_path, 'rb') as f:
            estimator = pickle.load(f)

        estimator = estimator.set_params(verbose=0, n_jobs=1)

        self.estimator = estimator

    def score(self, x):
        score = self.estimator.predict_proba([x])[0, 1]
        return score

    def score_batch(self, X):
        score = list(self.estimator.predict_proba(X)[:, 1])
        return score


def query_task_to_sql(query_task):

    sql = """
    SELECT *
    FROM indexed
    WHERE %s
    """ % ("\nAND ".join(["%s = '%s'" % (k, v) for k, v in query_task.items()]))

    return sql


def repl_func(match):
    return '*' * len(match.group(0))


NUMBER_LIKE_MASTER_REGEX = re.compile(
    r"(FLAT|UNIT|HOUSE|BUILDING|ROOM|BLOCK|BONDS?|FL|APARTMENT|\(F|F|-|\()? ?"
    r"((\d[A-Z]\d+)|(PF|BF|GF|[A-Z])?(\d+)([A-Z])?|(?<!')(^|\b)([A-Z]|GROUND)($|\b)(?!'))"
    r"\)?")


def strip_number_like(address):
    p = NUMBER_LIKE_MASTER_REGEX

    p_pc = re.compile(r'([A-Z]{1,2})([0-9][A-Z0-9]?)(?: *\n?)?([0-9])([A-Z])([A-Z])')

    # postcode = re.search(p_pc, address).group()

    address = re.sub(p_pc, repl_func, address)

    address = re.sub(p, repl_func, address)
    #
    # address = re.sub('&&', postcode, address)

    return address


def prepare_uprn(row):

    p = re.compile(r"(FLAT|UNIT|BUILDING|ROOM|BLOCK|BONDS?|FL|APARTMENT|\()? ?"
                   r"((\d[A-Z]\d+)|([A-Z]|PF|BF|GF)?(\d+)([A-Z])?|(?<!')(^|\b)([A-Z]|GROUND)(\b|$)(?!'))")

    columns = ['ORGANISATION_NAME', 'DEPARTMENT_NAME', 'SUB_BUILDING_NAME', 'BUILDING_NAME', 'BUILDING_NUMBER',
               'DEPENDENT_THOROUGHFARE', 'THOROUGHFARE', 'DOUBLE_DEPENDENT_LOCALITY', 'DEPENDENT_LOCALITY',
               'POST_TOWN']

    d_text = {col: re.sub(p, repl_func, row[col]) for col in columns}

    number_like = [row[col] for col in ['number_like_0', 'number_like_1', 'number_like_2',
                                        'number_like_3', 'number_like_4']]

    postcode_split = {col: row[col] for col in ['pc_area', 'pc_district', 'pc_sector', 'pc_unit_0', 'pc_unit_1']}

    return {'TEXTUAL': d_text, 'NUMBER_LIKE': number_like, 'POSTCODE_SPLIT': postcode_split}


def get_number_like_matching_matrix(s1, s2):

    mat = [0] * 25

    for i in range(5):
        for j in range(i, 5):
            if all([s1[i] == '', s2[j] == '']):
                mat[i * 5 + j] = 0
            elif s1[i] == '':
                mat[i * 5 + j] = -1
            elif s2[j] == '':
                mat[i * 5 + j] = -2
            elif s1[i] == s2[j]:
                mat[i * 5 + j] = 1
            else:
                mat[i * 5 + j] = -3

    return mat


def postcode_matching(split1, split2):
    return [int(split1[k] == split2[k]) for k in split1]


def summarize_features(nl, ts, pm):

    return nl + list(ts.values()) + pm


def guess_multiplier(max_level, max_flat, n_tenement):

    if max_flat > 1:

        if max_level * max_flat == n_tenement:
            return max_flat, 'exact'

        elif n_tenement % max_flat == 0:
            return max_flat, 'high_for_flat'

        else:
            return max_flat, 'medium_for_flat'

    elif max_level > 1:

        if n_tenement % max_level == 0:
            return int(n_tenement / max_level), 'high_for_level'

    else:

        if (n_tenement % 2 == 0) and (n_tenement % 3 == 0):

            return 2, 'low'

        elif n_tenement % 2 == 0:

            return 2, 'low'

        elif n_tenement % 3 == 0:

            return 3, 'low'

        elif n_tenement % 5 == 0:
            return 5, 'low'

    return 2, 'guess'



from zipfile import ZipFile
import csv
import io
import pandas as pd
import geopandas as gpd
import json
import os
import tqdm
import sqlite3
import re
import itertools

from flap.utils import flatten


class SqlDBManager:

    def __init__(self, project_name=None):
        self.global_db_path = os.path.join(os.getcwd(), 'db')
        self.project_name = project_name
        if project_name is not None:
            self.project_db_path = os.path.join(os.getcwd(), project_name, 'db')

    def list_global_db(self):
        db_names = os.listdir(self.global_db_path)
        return db_names

    def list_project_db(self):
        db_names = os.listdir(self.project_db_path)
        return db_names

    @staticmethod
    def create_db(path):
        new_db_path = path

        if os.path.exists(new_db_path):
            warn('DB:%s exists, new database cannot be created.' % db_name)
        else:
            os.makedirs(new_db_path, exist_ok=True)
            print('New db created at: %s' % new_db_path)
            os.mkdir(os.path.join(new_db_path, 'raw'))
            os.mkdir(os.path.join(new_db_path, 'sql'))
            os.mkdir(os.path.join(new_db_path, 'db_config'))

    def create_global_db(self, db_name):
        new_db_path = os.path.join(self.global_db_path, db_name)

        if os.path.exists(new_db_path):
            warn('DB:%s exists, new database cannot be created.' % db_name)
        else:
            os.makedirs(new_db_path, exist_ok=True)
            print('New db created at: %s' % new_db_path)
            os.mkdir(os.path.join(new_db_path, 'raw'))
            os.mkdir(os.path.join(new_db_path, 'sql'))
            os.mkdir(os.path.join(new_db_path, 'db_config'))

    def create_project_db(self, db_name):
        new_db_path = os.path.join(self.project_db_path, db_name)

        if os.path.exists(new_db_path):
            warn('DB:%s exists, new database cannot be created.' % db_name)
        else:
            os.makedirs(new_db_path, exist_ok=True)
            print('New db created at: %s' % new_db_path)
            os.mkdir(os.path.join(new_db_path, 'raw'))
            os.mkdir(os.path.join(new_db_path, 'sql'))
            os.mkdir(os.path.join(new_db_path, 'db_config'))

    def get_db_path(self, db_name, project_level=False):
        if not project_level:
            if db_name in self.list_global_db():
                return os.path.join(self.global_db_path, db_name)
            else:
                warn('DB: %s does not exist!' % db_name)
        else:
            if self.project_name is None:
                warn('Project not specified')
            else:
                if db_name in self.list_project_db():
                    return os.path.join(self.project_db_path, db_name)
                else:
                    warn('DB: %s does not exist!' % db_name)

    def get_db(self, db_name, project_level=False):

        return SqlDB(self.get_db_path(db_name, project_level))


class SqlDB:

    def __init__(self, path_to_db):
        self.db_path = path_to_db
        self.db_name = os.path.basename(self.db_path)
        self.sub_paths = {
            'raw': os.path.join(self.db_path, 'raw'),
            'db_config': os.path.join(self.db_path, 'db_config'),
            'sql_db': os.path.join(self.db_path, 'sql', 'db.sqlite3'),
            'sql_db_temp': os.path.join(self.db_path, 'sql', 'db_temp.sqlite3'),
            'vocabulary': os.path.join(self.db_path, 'vocabulary'),
            'pc1_mappings': os.path.join(self.db_path, 'vocabulary', 'pc1_mappings.json'),
            'pc1_mappings_region': os.path.join(self.db_path, 'vocabulary', 'pc1_mappings_region.json'),
            'unique_DOUBLE_DEPENDENT_LOCALITY': os.path.join(self.db_path, 'vocabulary',
                                                             'unique_DOUBLE_DEPENDENT_LOCALITY.json'),
            'unique_DEPENDENT_LOCALITY': os.path.join(self.db_path, 'vocabulary', 'unique_DEPENDENT_LOCALITY.json'),
            'unique_POST_TOWN': os.path.join(self.db_path, 'vocabulary', 'unique_POST_TOWN.json'),
            'thoroughfare_patterns': os.path.join(self.db_path, 'vocabulary', 'thoroughfare_patterns.json')
        }

        self.__db_status = None

    def __str__(self):
        self_str = f"""
        Database name: {self.db_name}
        Database path: {self.db_path}
        """
        return self_str

    def get_table_names(self):

        conn = self.get_conn()
        cur = conn.cursor()

        try:
            cur.execute("""SELECT name FROM sqlite_schema""")
            res = cur.fetchall()

            if len(res):
                table_names = [list(name)[0] for name in res]
            else:
                table_names = []

        except sqlite3.OperationalError:
            table_names = []

        cur.close()
        conn.close()

        return table_names

    def get_columns_of_table(self, table_name):

        table_name = (table_name,)

        conn = self.get_conn()
        cur = conn.cursor()
        cur.execute("""
        SELECT sql FROM sqlite_master WHERE tbl_name = ? AND type = 'table'
        """, table_name)
        res = cur.fetchall()

        s = res[0][0]
        lines = s.split('\n')
        column_names = [line.split('"')[1] for line in lines[1:-1]]

        return column_names

    @property
    def db_status(self):
        db_status = {
            'db_name': self.db_name,
            'db_path': self.db_path,
            'valid': os.path.isdir(self.db_path),
            'raw_added': len(os.listdir(self.sub_paths['raw'])) > 0
        }

        tables = self.get_table_names()

        db_status['table_raw_built'] = 'raw' in tables
        db_status['table_indexed_built'] = 'indexed' in tables

        self.__db_status = db_status

        return self.__db_status

    def drop_table(self, table):
        conn = self.get_conn()
        cur = conn.cursor()
        cur.execute(f"""DROP TABLE IF EXISTS {table}""")
        conn.commit()
        conn.close()

    def get_db_config(self):

        sql_config_path = [os.path.join(self.sub_paths['db_config'], file)
                           for file in os.listdir(self.sub_paths['db_config']) if '.json' in file]

        if len(sql_config_path):

            sql_config_file = sql_config_path[0]

            with open(sql_config_file, 'r') as f:
                sql_config = json.load(f)

        else:

            sql_config = {
                "RECORD_IDENTIFIER": "INTEGER",
                "CHANGE_TYPE": "TEXT",
                "PRO_ORDER": "TEXT",
                "UPRN": "TEXT",
                "UDPRN": "TEXT",
                "ORGANISATION_NAME": "TEXT",
                "DEPARTMENT_NAME": "TEXT",
                "SUB_BUILDING_NAME": "TEXT",
                "BUILDING_NAME": "TEXT",
                "BUILDING_NUMBER": "TEXT",
                "DEPENDENT_THOROUGHFARE": "TEXT",
                "THOROUGHFARE": "TEXT",
                "DOUBLE_DEPENDENT_LOCALITY": "TEXT",
                "DEPENDENT_LOCALITY": "TEXT",
                "POST_TOWN": "TEXT",
                "POSTCODE": "TEXT",
                "POSTCODE_TYPE": "TEXT",
                "DELIVERY_POINT_SUFFIX": "TEXT",
                "WELSH_DEPENDENT_THOROUGHFARE": "TEXT",
                "WELSH_THOROUGHFARE": "TEXT",
                "WELSH_DOUBLE_DEPENDENT_LOCALITY": "TEXT",
                "WELSH_DEPENDENT_LOCALITY": "TEXT",
                "WELSH_POST_TOWN": "TEXT",
                "PO_BOX_NUMBER": "TEXT",
                "PROCESS_DATE": "TEXT",
                "START_DATE": "TEXT",
                "END_DATE": "TEXT",
                "LAST_UPDATE_DATE": "TEXT",
                "ENTRY_DATE": "TEXT"
            }

            sql_config_file = os.path.join(self.sub_paths['db_config'], 'db_config.json')

            with open(sql_config_file, 'w') as f:
                json.dump(sql_config, f, indent=4)

            print('db_config created')
        return sql_config

    def build_raw(self, if_exist='skip'):
        db_status = self.db_status

        print(db_status)

        if not db_status['raw_added']:
            raise FileNotFoundError('Add raw database files to [db_path]/raw first and then run build')

        if if_exist == 'skip':

            if db_status['table_raw_built']:
                print('Table raw exists, build skipped. Use if_exist="replace", if you want to create a new db')
            else:
                _parsing_func(self)

        elif if_exist == 'replace':

            self.drop_table('raw')
            _parsing_func(self)

    def indexing_db(self, if_exist='skip'):
        db_status = self.db_status

        if if_exist == 'skip':

            if db_status['table_indexed_built']:
                print('TABLE indexed exists, build skipped. Use if_exist="replace", if you want to create a new db')
            else:
                _indexing_database(self)
                self.delete_temp()

        elif if_exist == 'replace':

            self.drop_table('indexed')
            _indexing_database(self)
            self.delete_temp()

    def build_vocabulary(self):

        if not os.path.exists(self.sub_paths['vocabulary']):
            os.mkdir(self.sub_paths['vocabulary'])

        compile_pc1_uniques(self)
        compile_pc1_mappings_region(self)
        compile_global_uniques(self)

        path_thoroughfare_patterns = os.path.join(os.getcwd(), 'src', 'parser', 'thoroughfare_patterns.json')
        os.system('cp %s %s' % (path_thoroughfare_patterns, self.sub_paths['thoroughfare_patterns']))

    def setup_database(self, if_exist='skip'):
        self.build_raw(if_exist=if_exist)
        self.build_vocabulary()
        self.indexing_db(if_exist=if_exist)

    def get_conn(self):
        return sqlite3.connect(self.sub_paths['sql_db'], timeout=10)

    def get_conn_temp(self):
        return sqlite3.connect(self.sub_paths['sql_db_temp'], timeout=10)

    def delete_temp(self):
        try:
            os.remove(self.sub_paths['sql_db_temp'])
        except FileNotFoundError:
            pass

    def sql_query(self, query):

        conn = self.get_conn()
        cur = conn.cursor()
        cur.execute(query)
        res = cur.fetchall()
        cur.close()
        conn.close()

        return res

    def sql_query_batch(self, query, batch_size=int(1e5)):
        conn = self.get_conn()
        cur = conn.cursor()
        cur.execute(query)

        for batch in iter(lambda: cur.fetchmany(batch_size), []):
            yield batch

        conn.close()

    def sql_table_batch_by_column(self, table_name, by_columns, batch_size, match_func=None):

        if isinstance(by_columns, str):
            by_columns = [by_columns]

        table_columns = self.get_columns_of_table(table_name)

        column_indices = [table_columns.index(col) for col in by_columns]

        query = """
            SELECT *
            FROM %s
            ORDER BY %s
        """ % (table_name, ','.join(by_columns))

        con = self.get_conn()
        cur = con.cursor()
        cur.execute(query)

        res = cur.fetchmany(batch_size)
        next_res = cur.fetchone()

        while len(res):

            try:

                if match_func is None:
                    while all([next_res[i] == res[-1][i] for i in column_indices]):
                        res.append(next_res)
                        next_res = cur.fetchone()
                else:
                    while match_func(next_res, res):
                        res.append(next_res)
                        next_res = cur.fetchone()

            except TypeError:
                pass

            res_df = pd.DataFrame.from_records(res)
            res_df.columns = table_columns

            yield res_df

            if next_res is None:
                break
            else:
                res = [next_res]
                res.extend(cur.fetchmany(batch_size))
                next_res = cur.fetchone()

        con.close()

    def create_index(self, table_name, columns):

        index_name = '__'.join(columns)

        sql = "CREATE INDEX IF NOT EXISTS %s ON %s(%s)" % (
            index_name,
            table_name,
            ', '.join(columns)
        )

        with self.get_conn() as conn:
            cur = conn.cursor()
            cur.execute(sql)


def _parsing_func(sql_db):
    zip_path = [os.path.join(sql_db.sub_paths['raw'], file)
                for file in os.listdir(sql_db.sub_paths['raw']) if '.zip' in file][0]

    db_file = sql_db.sub_paths['sql_db']

    conn = sqlite3.connect(db_file)

    with ZipFile(zip_path, 'r') as z:
        
        zz_files = [zz_file for zz_file in z.namelist()]

        if any('.gpkg' in zz_file for zz_file in zz_files):

            print('.gpkg database found. Starting to read db file')

            zz_data = io.BytesIO(z.read(zz_files[0]))

            gpkg_db = gpd.read_file(zz_data, layer='delivery_point_address')

            print('convert to object type')

            gpkg_db = convert_df_to_object(gpkg_db)

            cols_upper = [col.upper() for col in gpkg_db.columns]

            gpkg_db.columns = cols_upper

            print('Save to SQL database')

            gpkg_db.to_sql(name='raw', con=conn, if_exists='replace', dtype='TEXT', index=False)

            print('Raw table building finished')

        elif any('.zip' in zz_file for zz_file in zz_files):
            print('.zip database found. Starting to read db file')

            chunk_size = int(1e5)
            rows = []
            data_index = '28'

            sql_config = sql_db.get_db_config()

            zz_files = [zz_file for zz_file in z.namelist() if '.zip' in zz_file]

            for zz_file in tqdm.tqdm(zz_files):  
                zz_data = io.BytesIO(z.read(zz_file))

                with ZipFile(zz_data, 'r') as zz:
                    csv_files = [csv_file for csv_file in zz.namelist() if '.csv' in csv_file]

                    for csv_file in csv_files:

                        with zz.open(csv_file, 'r') as f:

                            lines = [line.decode() for line in f.readlines()]

                            reader = csv.reader(lines)

                            for row in reader:
                                if row[0] == data_index:
                                    rows.append(row)

                        while len(rows) > chunk_size:
                            df = pd.DataFrame.from_records(rows, columns=list(sql_config.keys()))
                            df.to_sql(name='raw', con=conn, if_exists='append', dtype=sql_config, index=False)
                            rows = rows[chunk_size:]

            df = pd.DataFrame.from_records(rows, columns=list(sql_config.keys()))
            df.to_sql(name='raw', con=conn, if_exists='append', dtype=sql_config, index=False)
            print('Raw table building finished')

        else:
            pass

    conn.close()


def compile_pc1_uniques(sql_db):
    batch_gen = sql_db.sql_table_batch_by_column(
        table_name='raw', by_columns='POSTCODE', batch_size=int(1e5),
        match_func=lambda next_res, res: next_res[15].split(' ')[0] == res[-1][15].split(' ')[0])

    columns_to_compile = ["THOROUGHFARE", "DOUBLE_DEPENDENT_LOCALITY", "DEPENDENT_LOCALITY", "POST_TOWN"]

    pc1_mappings = {}

    for col in columns_to_compile:
        pc1_mappings[col] = {}

    for df in tqdm.tqdm(batch_gen):

        df['pc1'] = df.apply(lambda row: row['POSTCODE'].split(' ')[0], axis=1)

        for pc1, group in df.groupby('pc1'):

            for col in columns_to_compile:

                uniques = group[col].unique().tolist()

                try:
                    uniques.remove('')
                except ValueError:
                    pass

                if len(uniques):
                    pc1_mappings[col][pc1] = uniques

    with open(sql_db.sub_paths['pc1_mappings'], 'w') as f:
        json.dump(pc1_mappings, f, indent=4)


def compile_pc1_mappings_region(sql_db):
    path_pc1_region_mappings = os.path.join(os.getcwd(), 'src', 'parser', 'pc1_mappings_region.json')
    os.system('cp %s %s' % (path_pc1_region_mappings, sql_db.sub_paths['pc1_mappings_region']))


def compile_global_uniques(sql_db):
    columns_to_compile = ["DOUBLE_DEPENDENT_LOCALITY", "DEPENDENT_LOCALITY", "POST_TOWN"]

    for col in columns_to_compile:
        res_sql = sql_db.sql_query("""
                                   SELECT DISTINCT %s
                                   FROM raw
                                   """ % col)
        res = list(itertools.chain(*res_sql))

        try:
            res.remove('')
        except ValueError:
            pass

        with open(sql_db.sub_paths['unique_%s' % col], 'w') as f:
            json.dump(res, f, indent=4)


def _indexing_database(sql_db):
    conn_temp = sql_db.get_conn_temp()

    tasks_gen = sql_db.sql_table_batch_by_column('raw', 'POSTCODE', int(1e5))

    for res in tqdm.tqdm(tasks_gen):
        res_indexed = indexing_uprn(res)
        res_indexed.to_sql(name='indexed',
                           con=conn_temp,
                           if_exists='append',
                           dtype='TEXT',
                           index=False)

    sql_chunk_gen = pd.read_sql(sql='SELECT * FROM indexed', con=conn_temp, chunksize=int(1e5))

    conn = sql_db.get_conn()

    for chunk in tqdm.tqdm(sql_chunk_gen):
        chunk.to_sql(name='indexed',
                     con=conn,
                     if_exists='append',
                     dtype='TEXT',
                     index=False
                     )

    conn.close()
    conn_temp.close()


def type_of_macro(row):

    cols = ['DEPENDENT_THOROUGHFARE', 'THOROUGHFARE', 'DOUBLE_DEPENDENT_LOCALITY', 'DEPENDENT_LOCALITY']

    return ''.join([str(int(len(row[col]) > 0)) for col in cols])


def type_of_micro(row):

    cols = ['ORGANISATION_NAME', 'DEPARTMENT_NAME', 'SUB_BUILDING_NAME',
            'BUILDING_NAME', 'BUILDING_NUMBER']

    return ''.join([str(int(len(row[col]) > 0)) for col in cols])


NUMBER_LIKE_MASTER_REGEX = re.compile(
    r"(FLAT|UNIT|BUILDING|ROOM|BLOCK|BONDS?|FL|PF|BF|GF|APARTMENT|\(F|F|-|\()? ?"
    r"((\dF\d+)|(\d+[A-Z]\d+)|([A-EG-Z])?(\d+)([A-Z])?|(?<!')(^|\b)([A-Z]|GROUND)($|\b)(?!'))"
    r"\)?")


def parse_number_like_uprn(row):

    p = NUMBER_LIKE_MASTER_REGEX

    split_pattern = re.compile(r"(?<=\d)(?=\D)|(?=\d)(?<=\D)")

    units = list()

    for col in ['SUB_BUILDING_NAME', 'BUILDING_NAME']:
        if len(row[col]):
            matches = list(re.finditer(p, row[col]))
            units.append(flatten([re.split(split_pattern, match.group(2))
             if re.search(r'\dF\d', match.group(2)) is None else [match.group(2)]
             for match in matches]))

    if len(row['BUILDING_NUMBER']):
        units.append([row['BUILDING_NUMBER']])

    res = flatten(units[::-1])

    if len(res) < 5:
        res += [''] * (5 - len(res))
    elif len(res) > 5:
        res = res[:5]

    return pd.Series(res)


def split_postcode_uprn(row):

    p = re.compile(r'([A-Z]{1,2})([0-9][A-Z0-9]?)(?: +)?([0-9])([A-Z])([A-Z])')

    match = re.match(p, row['POSTCODE'])

    pc0, pc1 = ''.join([match.group(i) for i in range(1, 3)]), ''.join([match.group(i) for i in range(3, 6)])

    pc_area, pc_district, pc_sector, pc_unit_0, pc_unit_1 = [match.group(i) for i in range(1, 6)]

    return pd.Series([pc0, pc1, pc_area, pc_district, pc_sector, pc_unit_0, pc_unit_1])


def count_n_tenement(df):

    df['n_tenement'] = 1

    groups = []

    for (i, j), group in df.groupby(['BUILDING_NAME', 'POSTCODE']):
        if (i != '') and (len(group) > 1):
            group['n_tenement'] = len(group)

        groups.append(group)

    df = pd.concat(groups)

    groups = []

    for (i, j), group in df.groupby(['BUILDING_NUMBER', 'POSTCODE']):
        if (i != '') and (len(group) > 1):
            group['n_tenement'] = len(group)
        groups.append(group)

    df = pd.concat(groups)

    return df


def indexing_uprn(df):

    df = count_n_tenement(df)

    df['type_of_micro'] = df.apply(type_of_micro, axis=1)
    df['type_of_macro'] = df.apply(type_of_macro, axis=1)

    df[['number_like_%s' % i for i in range(5)]] = df.apply(parse_number_like_uprn, axis=1)
    df[['pc0', 'pc1', 'pc_area', 'pc_district', 'pc_sector', 'pc_unit_0', 'pc_unit_1']] = \
        df.apply(split_postcode_uprn, axis=1)

    return df


# dm = SqlDBManager()
#
# dm.list_global_db()
#
# sql_db = dm.get_db(db_name='the_database')
#
# sql_db.setup_database(if_exist='replace')
#
# sql_db.sql_query("select * from indexed")
